/**
 * This algorithm is called Moore Neighbor Tracing
 * An explanation of the algorithm can be found here:
 * http://www.thebigblob.com/moore-neighbor-tracing-algorithm-in-c/
 *
 * @author Erik Smistad <smistad@idi.ntnu.no>
 */

#include "contour-tracing.h"
#include <SDL/SDL.h>

// The filename of the bmp image we want to run the algorithm on
const char * filename = "images/lazy-circle.bmp";
bool running = true;

/* Keyboard callback */
static void handleKeyDown(SDL_Event * event)
{
	switch (event->key.keysym.sym)
	{
	case SDLK_ESCAPE:
	case SDLK_q:
	case SDLK_x:
		running = false;
		break;
	}
}

/* General event callback */
static void processEvents()
{
	SDL_Event event;

	// Grab all the events off the queue
	while (SDL_PollEvent(&event))
	{
		switch (event.type)
		{
		case SDL_KEYDOWN:
			handleKeyDown(&event);
			break;
		case SDL_QUIT:
			// quit requests (like Ctrl-c)
			running = false;
			break;
		}
	}
}

int main(int argc, char **argv)
{
	// initialize SDL's video subsystem
	if (SDL_Init(SDL_INIT_VIDEO) < 0)
	{
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
	}
	// retrieve video information
	const SDL_VideoInfo *info = SDL_GetVideoInfo();
	if (!info)
	{
		fprintf(stderr, "Video query failed: %s\n", SDL_GetError());
		exit(-1);
	}

	// Load image
	SDL_Surface * image = SDL_LoadBMP(filename);
	if (!image)
	{
		fprintf(stderr, "Couldn't load %s: %s\n", filename, SDL_GetError());
		exit(-1);
	}

	// Set video mode
	SDL_Surface * screen = SDL_SetVideoMode(image->w, image->h,
			info->vfmt->BitsPerPixel, SDL_SWSURFACE);
	if (!screen)
	{
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
		exit(-1);
	}

	// Copy loaded image to surface
	SDL_Rect area = {0, 0, image->w, image->h};
	SDL_BlitSurface(image, NULL, screen, &area);
	SDL_FreeSurface(image);
	SDL_UpdateRect(screen, 0, 0, 0, 0);
	SDL_SaveBMP(screen, "images/before.bmp");

	SDL_Surface * borderImageSurface = SDL_CreateRGBSurface(
			SDL_SWSURFACE,
			screen->w,
			screen->h,
			info->vfmt->BitsPerPixel,
			info->vfmt->Rmask,
			info->vfmt->Gmask,
			info->vfmt->Bmask,
			info->vfmt->Amask);

	// Run the algorithm, show it on screen and save it
	borderImageSurface->pixels = mooreNeighborTracing((pixel *)screen->pixels, screen->w, screen->h);
	SDL_BlitSurface(borderImageSurface, NULL, screen, &area);
	SDL_SaveBMP(screen, "images/after.bmp");

	// Redraw screen
	SDL_UpdateRect(screen, 0, 0, 0, 0);

	// Main Loop
	while (running)
	{
		// Process events
		processEvents();

		// refresh 10 frames per second
		SDL_Delay(10/1000);
	}

	// Unload SDL
	SDL_Quit();
	return 0;
}
